/**
 * @file acli_api.h
 *
 * @brief aCli API definition.
 *
 **/
#ifndef _ACLI_API_H_
#define _ACLI_API_H_


#ifdef __cplusplus
extern "C" {
#endif


/**
 * @brief Names of predefined media and modes
 */
/** @{*/
#define ACLI_MEDIA_NAME_STDIO       "console"
#define ACLI_MEDIA_NAME_TELNET      "telnet"
#define ACLI_MODE_NAME_DEMO         "demo"
/** @}*/


/** 
 * @enum ACLI_MEDIA
 * @brief List of supported media types.
 */
typedef enum
{
    ACLI_MEDIA_STREAM   = 0x01,     /**< regular stream */
    ACLI_MEDIA_NET      = 0x04      /**< socket connection */
} ACLI_MEDIA;


/** 
 * @enum ACLI_MODE
 * @brief List of supported mode types.
 */
typedef enum
{
    ACLI_MODE_REGULAR   = 0x01,     /**< regular mode */
} ACLI_MODE;


/**
 * @struct _ACLI_CREDENTIAL
 * @brief Authentification information
 */
typedef struct _ACLI_CREDENTIAL
{
    char      username[10];               /**< username */
    char      password[10];               /**< password */
} ACLI_CREDENTIAL;


/** 
 * @enum ACLI_SHELL_STATE_RESPOND
 * @brief Customer is able to set respond function on entering shell state. In this list
 *        possible responses are enumerated
 */
typedef enum
{
    ACLI_SHELL_RESPOND_OPEN   = 0x0,    /**< respond on session opening */
    ACLI_SHELL_RESPOND_CLOSE,           /**< respond on session closing */
    ACLI_SHELL_RESPOND_MAX              /**< max number of supported responses (should start from 0 to max) */
} ACLI_SHELL_RESPOND;


/**
 * @struct _ACLI_SHELL_INFO
 * @brief This type is container for information describing shell
 */
typedef struct _ACLI_SHELL_INFO
{
    union
    {
        char    addr_in[4];
    } media_info;       
} ACLI_SHELL_INFO;


/**
 * @struct _ACLI_CMD_DESC
 * @brief This structure can be used to define command (keywords, arguments, help and callback)
 */
typedef struct _ACLI_CMD_DESC
{
    const char*     cmd;
    const char*     desc;
    const char*     (*proc)(int, const void **, void *);
    void*           context;
} ACLI_CMD_DESC;


/**
 * @union _ACLI_OBJ_EXT
 * @brief Set additional parameters that allow to configure media or layout objects
 */
typedef union _ACLI_OBJ_EXT
{
    /* media ACLI_MEDIA_STREAM parameters */
    struct
    {
        char*   in_file_name;           /**< file name for input */
        char*   out_file_name;          /**< file name for output */
        int     cache;                  /**< sets the cache to be used for I/O operations with the specified 
                                             stream, which becomes a fully buffered stream, or alternatively, 
                                             if the argument for buffer is NULL, it disables buffering for 
                                             the stream, which becomes an unbuffered stream */
    } stream;       

    /* media ACLI_MEDIA_NET parameters */
    struct
    {
        short   port;                   /**< server port */
        int     cache;                  /**< sets the cache to be used for I/O operations with the specified 
                                             stream, which becomes a fully buffered stream, or alternatively, 
                                             if the argument for buffer is NULL, it disables buffering for 
                                             the stream, which becomes an unbuffered stream */
        char*   handshake;              /**< extra configuration setting (it can be initial protocol settings)*/
    } socket;       

    /* mode parameters */
    struct
    {
        char*               prompt;         /**< prompt of the mode */
        char*               greeting;       /**< shell`s greeting */
        ACLI_CREDENTIAL*    auth;           /**< user credentials */
        ACLI_CMD_DESC*      cmd_desc_tbl;   /**< command description table where the latest element 
                                                 should be zero to demonstrate the end of data*/
    } mode;       
} ACLI_OBJ_EXT;


/**
 * @struct _ACLI_OBJ
 * @brief Describes component objects
 */
typedef struct _ACLI_OBJ
{
    const char*     name;               /**< component name */
    int             id;                 /**< component id */
    ACLI_OBJ_EXT*   context;            /**< component extended information */
} ACLI_OBJ;


/**
 * @struct _ACLI_FORM
 * @brief This type allows to define initial configuration.
 */
typedef struct _ACLI_FORM
{
    ACLI_OBJ*       media_array;        /**< array with media description information
                                             where the latest element should be zero to
                                             demonstrate the end of data */
    ACLI_OBJ*       mode_array;        /**< array with mode description information
                                             where the latest element should be zero to
                                             demonstrate the end of data */
} ACLI_FORM;


/**
 * acli_init
 *
 * @brief
 *    The function is used to initialize internal object.
 *    This function should be called at least once before using any other function.
 *    Special structured data can be passed to set initial logging parameters.
 *    In case user defined configuartion is not passed logging sets default parameters.
 *
 * @param[in]    config         Initial configuartion.
 *
 * @retval zero - on success
 * @retval non-zero - on failure
 ***************************************************************************/
extern int acli_init( ACLI_FORM* config );


/**
 * acli_exit
 *
 * @brief
 *    The function is used as a destructor.
 *
 * @retval zero - on success
 * @retval non-zero - on failure
 ***************************************************************************/
extern int acli_exit( void );


/**
 * acli_create
 *
 * @brief
 *    Create new shell and identify one with unique name.
 *
 * @param[in]    name           Shell name.
 *
 * @retval zero - on success
 * @retval non-zero - on failure
 ***************************************************************************/
extern int acli_create( const char* name );


/**
 * acli_run
 *
 * @brief
 *    The function runs CLI shell. Mode and media should exist for the shell.
 *    If the argument is a null pointer, all existing shells are run.
 *
 * @param[in]    name           Shell name.
 *
 * @retval zero - on success
 * @retval non-zero - on failure
 ***************************************************************************/
extern int acli_run( const char* name );


/**
 * acli_quit
 *
 * @brief
 *    The function quits CLI shell.
 *    If the argument is a null pointer, all existing shells are run.
 *
 * @param[in]    name           Shell name.
 *
 * @retval zero - on success
 * @retval non-zero - on failure
 ***************************************************************************/
extern int acli_quit( const char* name );


/**
 * acli_set_mode
 *
 * @brief
 *    Switch the shell to new mode.
 *
 * @param[in]    name           Shell name.
 * @param[in]    mode_name      Mode name.
 *
 * @retval zero - on success
 * @retval non-zero - on failure
 ***************************************************************************/
extern int acli_set_mode( const char* name,
                          const char* mode_name );


/**
 * acli_add_capability
 *
 * @brief
 *    Set default media and mode for the shell.
 *
 * @param[in]    name           Shell name.
 * @param[in]    media_name     Media name.
 * @param[in]    mode_name      Mode name.
 *
 * @retval zero - on success
 * @retval non-zero - on failure
 ***************************************************************************/
extern int acli_add_capability( const char* name,
                                const char* media_name,
                                const char* mode_name);


/**
 * acli_set_respond
 *
 * @brief
 *    Set customer response to transition into new shell state.
 *
 * @param[in]    name           Shell name.
 * @param[in]    id             Respond identificator.
 * @param[in]    proc           Respond procedure.
 * @param[in]    context        Respond context.
 *
 * @retval zero - on success
 * @retval non-zero - on failure
 ***************************************************************************/
extern int acli_set_respond( const char* name,
                             ACLI_SHELL_RESPOND id,
                             void (*proc)(void *),
                             void *context);


/**
 * acli_get_info
 *
 * @brief
 *    Retrieve information about shell.
 *
 * @param[in]   name          Shell name.
 * @param[out]  info          Shell info content.
 *
 * @retval zero - on success
 * @retval non-zero - on failure
 ***************************************************************************/
extern int acli_get_info( const char* name,
                          ACLI_SHELL_INFO *info);


/**
 * acli_strerr
 *
 * @brief
 *    The function maps errnum to an error-message string, returning a pointer to the string.
 *
 * @param[in]    errcode         Error number.
 *
 * @retval pointer to error string message - on success
 * @retval NULL - on failure
 ***************************************************************************/
extern const char* acli_strerr( int errcode );


#ifdef __cplusplus
}
#endif


#endif /* _ACLI_API_H_ */

