/*
 * Copyright (C) Voltaire Ltd. 2010-2011.  ALL RIGHTS RESERVED.
 * This software product is a proprietary product of Voltaire Ltd.
 * (the "Company") and all right, title, and interest and to the software product,
 * including all associated intellectual property rights, are and shall
 * remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 * $COPYRIGHT$
 * $HEADER$
 */

#ifndef __PJOIN_API_H__
#define __PJOIN_API_H__

/**
 * @file pjoin_api.h
 *
 * @brief API for multicast proxy join
 */

#include <stdlib.h>
#include <inttypes.h>
#include "vlt_buffer.h"
#include "pjoin_settings.h"

typedef struct mgid_s {
    uint8_t raw[16];
} pjoin_mgid_t;
typedef pjoin_mgid_t pjoin_gid_t;

typedef struct mgid_signature_s {
    uint8_t raw[2];
} pjoin_mgid_signature_t;

typedef struct _mc_key_t{
    uint64_t sakey;
    pjoin_mgid_signature_t sign;
    uint32_t qkey;
} mc_key_t;

typedef uint64_t pjoin_guid_t;

/**
 * @brief Join list of guids to existing multicast group or create a new one
 *
 * This function takes list of guids of num_of_guids size and join them to multicast
 * group (*mlid, *mgid). If (*mlid, *mgid) is 0 new multicast group will be created 
 * and its mlid, mgid will be set to the (*mlid, *mgid) variables. 
 *
 * @param[in, out] mlid Pointer to multicast group LID 
 * @param[in, out] mgid Pointer to multicast grou GID
 * @param[in] num_of_guids Number of GUIDs to join multicast group
 * @param[in] guids List of GUIDs of num_of_guids size to join multicast group
 * @param[in] footer Content of "globals" section which will be returned in pjoin_get_phys_tree
 * @param[in] join_icpus If 1 icpu switches will join multicast group 
 *
 * @retval ::ErrorCode error code
 */
int pjoin_join(uint16_t *mlid, 
               pjoin_mgid_t *mgid, 
               int num_of_guids, 
               pjoin_guid_t *guids, 
               char *footer, 
               int join_icpus);

/**
 * @brief Leave list of guids from multicast group
 *
 * @param[in] mlid Multicast group LID 
 * @param[in] mgid Multicast group GID
 * @param[in] num_of_guids Number of GUIDs to leave multicast group
 * @param[in] guids List of GUIDs of num_of_guids size to leave multicast group
 *
 * @retval ::ErrorCode error code
 */
int pjoin_leave(uint16_t mlid, 
                pjoin_mgid_t *mgid, 
                int num_of_guids, 
                pjoin_guid_t *guids);

/**
 * @brief Get physical or logical tree in yaml format
 * 
 * @param[in] mlid Multicast group LID
 * @param[in] mgid Multicast group GID
 * @param[in] optimize Whether or not to optimize multicast tree
 * @param[out] yaml Yaml string with tree description
 *
 * @retval ::ErrorCode error code
 */
int pjoin_get_phys_tree(uint16_t mlid, 
                        pjoin_mgid_t *mgid, 
                        int optimize, 
                        vlt_buffer *yaml);

/**
 * @brief Init libvtopo
 *
 * The init function takes device name and port number which will be used by default to communicate with OpenSM.
 * In case those are nulls first active port will be used
 *
 * @param[in] dev_name Device name to communicate with OpenSM
 * @param[in] port_num Port number to communicate with OpenSM
 * @param[in] mkey     Management key to communicate with OpenSM
 *
 * @retval ::ErrorCode error code
 */
int pjoin_init(char *dev_name, 
               int port_num,
               uint64_t mkey);

/**
 * @brief Uninitialize libvtopo
 *
 * The unitialize function should be called last to free memory 
 *
 * @retval ::ErrorCode error code
 */
int pjoin_done();

/**
 * @brief Sweep network topology via MAD messages
 *
 * @retval ::ErrorCode error code
 */
int pjoin_sweep_topology();

/**
 * @brief Sweep topology fro file in ibnetdiscover format 
 *
 * @param[in] filename path to ibnetdiscover topology description
 *
 * @retval ::ErrorCode error code
 */
int pjoin_sweep_topology_from_file(char *filename);

/**
 * @brief Update list of iCPU guids 
 *
 * @param[in] guids Array of iCPU switch GUIDs
 * @param[in] num_of_guids Number of GUIDs
 *
 * @retval ::ErrorCode error code
 */
int pjoin_update_icpu_switches(pjoin_guid_t *guids, 
                               int num_of_guids);

/**
 * @brief Calculate number of switches with spsific vendor id
 *
 * @param[in] id Vendor id
 * @param[out] num_of_switches Number of switches with spesific vendor id
 *
 * @retval ::ErrorCode error code
 */
int pjoin_check_switch_vendid(uint32_t id,
                              int *num_of_switches);

/**
 * @brief Update logical tree weight estimating and optimization settings
 *
 * @param[in] settings Logical tree weight estimating and optimization settings
 *
 * @retval ::ErrorCode error code
 */
int pjoin_set_settings(PjoinSettings settings);

/**
 * @brief Clean multicast groups whose MGID matchs the pattern (that is group is created by libvtopo)
 * @param[in] sign Multicast Group's signature to filter
 *
 * @retval ::ErrorCode error code
 */
int pjoin_cleanup(mc_key_t mc_key);

#endif

