/**
 * Copyright (C) Voltaire Ltd. 2001-2010.  ALL RIGHTS RESERVED.
 * This software product is a proprietary product of Voltaire Ltd.
 * (the "Company") and all right, title, and interest and to the software product,
 * including all associated intellectual property rights, are and shall
 * remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 * $COPYRIGHT$
 * $HEADER$
 */

#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mpi.h>

#define USEC_PER_SEC	1000000L

struct test_dtype {
	const char *name;
	MPI_Datatype dtype;
	size_t size;
};

struct test_op {
	const char *name;
	MPI_Op op;
};

struct test_dtype test_dtypes[] = {
	{ "MPI_BYTE",   MPI_BYTE,          sizeof(char) },
	{ "MPI_INT",    MPI_INT,           sizeof(int) },
	{ "MPI_DOUBLE", MPI_DOUBLE,        sizeof(double) },
	{ "MPI_FLOAT",  MPI_FLOAT,         sizeof(float) },
	{ "MPI_LONG",   MPI_LONG,          sizeof(long) },
	{ NULL,         MPI_DATATYPE_NULL, 0 }
};

struct test_op test_ops[]  = {
	{ "MPI_SUM", MPI_SUM     },
	{ "MPI_MAX", MPI_MAX     },
	{ "MPI_BXOR", MPI_BXOR   },
	{ NULL,      MPI_OP_NULL }
};

struct test_spec {
	MPI_Comm         comm;
	MPI_Datatype     dtype;
	MPI_Op           op;
	void             *src;
	void             *dst;
	size_t           length;
	int              root;
	int              iternum;
};

typedef void (*test_coll_func_t)(struct test_spec *spec);

struct test_coll {
	const char       *name;
	test_coll_func_t func;
	int              using_dtype;
	int              using_op;
	int              using_size;
};

#define test_print(rank, fmt, ...) \
	if (rank == 0) { \
		printf(fmt "\n", ## __VA_ARGS__); \
	}

void test_coll_bcast(struct test_spec *spec)
{
	MPI_Bcast(spec->dst, spec->length, spec->dtype, spec->root, spec->comm);
}

void test_coll_reduce(struct test_spec *spec)
{
	MPI_Reduce(spec->src, spec->dst, spec->length, spec->dtype, spec->op,
	           spec->root, spec->comm);
}

void test_coll_allreduce(struct test_spec *spec)
{
	MPI_Allreduce(spec->src, spec->dst, spec->length, spec->dtype, spec->op,
	              spec->comm);
}

void test_coll_barrier(struct test_spec *spec)
{
	MPI_Barrier(spec->comm);
}

void test_coll_allgather(struct test_spec *spec)
{
	static int count;
	int size;

	if (spec->iternum == 0) {
		MPI_Comm_size(spec->comm, &size);
		count = spec->length / size;
	}

	MPI_Allgather(spec->src, count, spec->dtype, spec->dst, count,
	              spec->dtype, spec->comm);
}

void test_coll_allgatherv(struct test_spec *spec)
{
	static int count;
	static int *rcounts = NULL;
	static int *displs = NULL;
	static int scount;
	int rank, pos, i, size;

	if (spec->iternum == 0) {
		free(rcounts);
		free(displs);

		MPI_Comm_size(spec->comm, &size);
		MPI_Comm_rank(spec->comm, &rank);

		/* Arrange that all ranks together will fill the buffer */
		rcounts = malloc(size * sizeof *rcounts);
		displs = malloc(size * sizeof *displs);
		for (i = 0; i < size; ++i) {
			displs[i] = (spec->length * i) / size;
		}
		for (i = 0; i < size - 1; ++i) {
			rcounts[i] = displs[i + 1] - displs[i];
		}
		rcounts[size - 1] = spec->length - displs[size - 1];
		scount = rcounts[rank];
	}

	MPI_Allgatherv(spec->src, scount, spec->dtype, spec->dst, rcounts, displs,
	               spec->dtype, spec->comm);
}

struct test_coll test_colls[] = {
	{ "Bcast", test_coll_bcast, 1, 0, 1 },
	{ "Barrier", test_coll_barrier, 0, 0, 0 },
	{ "Reduce", test_coll_reduce, 1, 1, 1 },
	{ "Allreduce", test_coll_allreduce, 1, 1, 1 },
	{ "Allgather", test_coll_allgather, 1, 0, 1 },
	{ "Allgatherv", test_coll_allgatherv, 1, 0, 1 },
	{ NULL, NULL, 0, 0 }
};


void inline test_print_perf(int rank, const char *title, size_t nloop,
                            size_t *nloop_prev, struct timeval *tv_first,
                            struct timeval *tv_prev, size_t *interval,
                            int final)
{
	struct timeval tv_current, tv_diff;
	double d_num_usec, d_latency;
	double total_num_usec, avg_latency;

	if ((nloop % *interval != 0) && !final)
		return;

	gettimeofday(&tv_current, NULL);
	if (!timerisset(tv_first)) {
		*tv_first = tv_current;
	} else {
		timersub(&tv_current, tv_prev, &tv_diff);
		d_num_usec = tv_diff.tv_usec + tv_diff.tv_sec * USEC_PER_SEC;
		d_latency = d_num_usec / (nloop - *nloop_prev);
		
		/* Adjust interval to 1/2 second */
		if (d_latency > 0.5)
			*interval = ((USEC_PER_SEC / d_latency) / 2) + 1;
		else
			*interval = USEC_PER_SEC;
		
		if (tv_diff.tv_sec == 0 && !final)
			return;

		timersub(&tv_current, tv_first, &tv_diff);
		total_num_usec = tv_diff.tv_usec + tv_diff.tv_sec * USEC_PER_SEC;
		avg_latency = total_num_usec / nloop;

		if (final) {
			test_print(rank, "%s: %.2f us", title, avg_latency);
		} else {
			test_print(rank, "[%8lu] %s: %.2f us (avg: %.2f us)",
			           nloop, title, d_latency, avg_latency);
		}
	}
	*tv_prev = tv_current;
	*nloop_prev = nloop;
}

void test_run(const char *title, struct test_dtype *dtype, struct test_op *op,
              size_t nloops, size_t nbytes, int root, test_coll_func_t coll_func)
{
	struct test_spec spec;
	struct timeval tv_first, tv_prev;
	size_t nloop, nloop_prev;
	size_t interval;
	int rank, comm_size;

	spec.src = malloc(nbytes);
	spec.dst = malloc(nbytes);
	if (!spec.src || !spec.dst) {
		fprintf(stderr, "Failed to allocate memory: %m\n");
		return;
	}

	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &comm_size);

	memset(spec.src, 0, nbytes); // TODO put meaningful values
	memset(spec.dst, 0, nbytes);

	if (dtype->size != 0) {
		spec.length = nbytes / dtype->size;
		if (spec.length == 0)
			return;

		if (spec.length * dtype->size != nbytes) {
			test_print(rank,
			           "Warning: '%s' is not filling entire buffer",
				   dtype->name);
		}
	}

	timerclear(&tv_first);
	timerclear(&tv_prev);
	nloop_prev = 0;
	interval = 1;

	spec.comm = MPI_COMM_WORLD;
	spec.dtype = dtype->dtype;
	spec.op = op->op;
	spec.root = root == -1 ? 0 : root;

	MPI_Barrier(spec.comm);
	for (nloop = 0; nloop < nloops; ++nloop) {
//		memset(srcbuf, 0, nbytes);
//		memset(dstbuf, 0, nbytes);
		spec.iternum = nloop;
		coll_func(&spec);
//		memcpy(srcbuf, dstbuf, nbytes);
		test_print_perf(rank, title, nloop, &nloop_prev, &tv_first,
		                &tv_prev, &interval, 0);
		spec.root = root == -1 ? (spec.root + 1) % comm_size : root;
	}
	MPI_Barrier(spec.comm);

	test_print_perf(rank, title, nloop, &nloop_prev, &tv_first, &tv_prev,
	                &interval, 1);

	free(spec.dst);
	free(spec.src);
}

void test_make_title(struct test_dtype *dtype, struct test_op *op,
                     struct test_coll *coll, size_t nbytes, char *buf,
                     size_t maxlen)
{
	char *p = buf, *end = buf + maxlen - 1;
	buf[0] = 0;
	snprintf(p, end - p, "%s ", coll->name);
	p += strlen(p);

	if (coll->using_size) {
		snprintf(p, end - p, "%lu bytes, ", nbytes);
		p += strlen(p);
	}

	if (coll->using_dtype) {
		snprintf(p, end - p, "%s ", dtype->name);
		p += strlen(p);
	}

	if (coll->using_op) {
		snprintf(p, end - p, "%s ", op->name);
		p += strlen(p);
	}

	*p = 0;
}

static int test_match_filter(const char *filter, const char *name)
{
	char *dup, *p;
	int found;

	if (!filter)
		return 1;

	dup = strdup(filter);
	p = strtok(dup, ",");
	found = 0;
	while (p && !found) {
		found |= !strcasecmp(name, p);
		p = strtok(NULL, ",");
	}
	free(dup);
	return found;
}

static void test_run_all(size_t nloops, size_t nbytes, int root, char *coll_filter,
                         char *dtype_filter, char *op_filter)
{
	struct test_dtype *dtype;
	struct test_op *op;
	struct test_coll *coll;
	char title[256], *p;
	int rank;

	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	for (coll = test_colls; coll->name; ++coll) {
		if (!test_match_filter(coll_filter, coll->name))
			continue;
		for (dtype = test_dtypes; dtype->name; ++dtype) {
			if (coll->using_dtype && !test_match_filter(dtype_filter, dtype->name))
				continue;
			for (op = test_ops; op->name; ++op) {
				if (coll->using_op && !test_match_filter(op_filter, op->name))
					continue;
				test_make_title(dtype, op, coll, nbytes, title, sizeof(title));
				test_run(title, dtype, op, nloops, nbytes, root, coll->func);
				if (!coll->using_op)
					break;
			}
			if (!coll->using_dtype)
				break;
		}
	}
}

void test_help(const char *progname)
{
	struct test_dtype *dtype;
	struct test_op *op;
	struct test_coll *coll;

	fprintf(stderr,
	        "Usage: %s <num_loops> <num_bytes> <root> [ <coll> ] [ < dtype> ] [ <op> ]\n",
	        progname);

	fprintf(stderr, "  coll:  ");
	for (coll = test_colls; coll->name; ++coll)
		fprintf(stderr, "%s ", coll->name);
	fprintf(stderr, "\n");

	fprintf(stderr, "  dtype: ");
	for (dtype = test_dtypes; dtype->name; ++dtype)
		fprintf(stderr, "%s ", dtype->name);
	fprintf(stderr, "\n");

	fprintf(stderr, "  op:    ");
	for (op = test_ops; op->name; ++op)
		fprintf(stderr, "%s ", op->name);
	fprintf(stderr, "\n");
}

int main(int argc, char* argv[])
{
	size_t nloops, nbytes;
	int rank, root;

	MPI_Init(&argc, &argv);
	if (argc < 4) {
		test_help(argv[0]);
		return -1;
	}

	nloops = atoi(argv[1]);
	nbytes = atoi(argv[2]);
	root   = atoi(argv[3]);

	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	test_print(rank, "-------------- Test Start--------------");
	test_run_all(nloops, nbytes, root,
	             argc >= 4 ? argv[4] : NULL,
	             argc >= 5 ? argv[5] : NULL,
	             argc >= 6 ? argv[6] : NULL);
	test_print(rank, "--------------- Test End --------------");

	MPI_Finalize();
	return 0;
}
